import logging
import sys
import threading

from arrnounced import irc
from arrnounced import webui

from arrnounced.tracker import Tracker, TrackerConfig
from arrnounced.tracker_xml_config import get_tracker_xml_configs

logger = logging.getLogger("MANAGER")


def _get_trackers(user_config, tracker_config_path):
    xml_configs = get_tracker_xml_configs(tracker_config_path)
    trackers = {}
    for user_tracker in user_config.trackers:
        if user_tracker.type not in xml_configs:
            logger.error(
                "Tracker '%s' from configuration is not supported", user_tracker.type
            )
        elif _are_settings_configured(
            user_tracker, xml_configs[user_tracker.type].settings
        ):
            trackers[user_tracker.type] = Tracker(
                TrackerConfig(user_tracker, xml_configs[user_tracker.type])
            )
    return trackers


# Check that all setting from the XML tracker config is configured in the user config.
def _are_settings_configured(user_tracker, required_settings):
    configured = True
    for setting in required_settings:
        if setting == "cookie":
            logger.warning(
                "%s: Tracker seems to require cookies to download torrent files. "
                + "Sonarr/Radarr/Lidarr API does not support cookies",
                user_tracker.type,
            )
        elif setting not in user_tracker.settings:
            logger.error("%s: Must specify '%s' in config", user_tracker.type, setting)
            configured = False
    return configured


def run(user_config, tracker_config_path):
    trackers = _get_trackers(user_config, tracker_config_path)
    if len(trackers) == 0:
        logger.error("No trackers configured, exiting...")
        sys.exit(1)

    irc_thread = threading.Thread(target=irc.run, args=(trackers,))
    webui_thread = threading.Thread(target=webui.run, args=(user_config,))

    irc_thread.start()
    webui_thread.start()

    irc_thread.join()
    webui_thread.join()

    logger.debug("Threads joined")



import pydle
import re

# 004 format
# <server_name> <version> <user_modes> <chan_modes> <channel_modes_with_params> <user_modes_with_params> <server_modes> <server_modes_with_params>
# 004 example
# :d7574991db48.example.com 4 nickname d7574991db48.example.com InspIRCd-3 BIRWcgikorsw ACHIKMOPRTXabcefghijklmnopqrstvz HIXabefghjkloqv

regex_004 = re.compile(r":[^\s]+ \d+ ([^\s]+ ){3}([\w]+ ){2}(?P<param_modes>[\w]+)")
regex_mode = re.compile(r"[-+](?P<modes>\w+)")
rfc1459_param_modes = "bovkl"


# This class is created purely because pydle throws uncatchable exceptions on
# modes with parameters it does not know about. An extension to the IRC
# protocol specifies all modes with parameters the server uses. As far I
# as can tell the extension is not part of any standard but seems to be widely
# used.
# channel_modes_with_params is the focus of fixing
class ModesFixer(pydle.features.TLSSupport):
    def __init__(self, *args, **kwargs):
        self.custom_modes = []
        super().__init__(*args, **kwargs)

    # Store all parameter modes which is not part of RFC1459. RFC1495 is what
    # pydle mainly implements.
    async def on_raw_004(self, message):
        str_message = str(message)
        match = regex_004.match(str_message)
        if match:
            for mode in match.group("param_modes"):
                if mode not in rfc1459_param_modes:
                    self.custom_modes.append(mode)
        await super().on_raw_004(message)

    # TODO: This function currently depends on the internal workings of pydle.
    def _clean_modes(self, message, modes):
        removed = 0
        for i, mode in enumerate(modes):
            if mode in self.custom_modes:
                # Removed mode from parameter
                message.params[1] = message.params[1].replace(mode, "")
                # Then remove the parameter of the mode
                del message.params[2 + i - removed]
                removed = removed + 1

        # All modes removed
        if removed == len(modes):
            return None

        return message

    def _find_and_clean_modes(self, message):
        match = regex_mode.match(message.params[1])
        if match:
            modes = match.group("modes")
            message = self._clean_modes(message, modes)

        return message

    async def on_raw_324(self, message):
        message = self._find_and_clean_modes(message)
        if message is None:
            return

        await super().on_raw_324(message)

    # Cleanse any non RFC1495 parameter modes from the message before forwarding
    # it to pydle.
    async def on_raw_mode(self, message):
        if self.is_channel(message.params[0]):
            message = self._find_and_clean_modes(message)

        if message is None:
            return

        await super().on_raw_mode(message)



import asyncio
import html
import logging
from pony.orm.core import TransactionError

from arrnounced import announce_parser
from arrnounced import db
from arrnounced import utils
from arrnounced.announcement import create_announcement
from arrnounced.backend import notify, notify_which_backends

logger = logging.getLogger("MESSAGE_HANDLER")


def _is_announcement(source, target, tracker):
    return (
        source in tracker.config.announcer_names
        and target in tracker.config.irc_channels
    )


def _sanitize_message(message):
    message = utils.strip_irc_color_codes(message)
    message = html.unescape(message)
    return message


def _check_and_parse(tracker, source, target, message):
    if not _is_announcement(source, target, tracker):
        logger.debug("Message is no announcement")
        return

    message = _sanitize_message(message)

    variables = announce_parser.parse(tracker, message)
    if variables is None:
        return None

    return create_announcement(tracker, variables)


async def _handle_announcement(tracker, announcement):
    backends = notify_which_backends(tracker, announcement.category)

    backends_string = (
        "/".join([b.name for b in backends]) if len(backends) > 0 else "None"
    )

    if tracker.config.announce_delay > 0:
        logger.debug(
            "%s: Waiting %s seconds to notify %s",
            tracker.config.short_name,
            tracker.config.announce_delay,
            backends_string,
        )
        await asyncio.sleep(tracker.config.announce_delay)

    with db.db_session:
        db_announced = db.insert_announcement(announcement, backends_string)
    logger.info(
        "Notifying %s of release from %s: %s",
        backends_string,
        tracker.config.short_name,
        announcement.title,
    )

    backend = await notify(announcement, backends)

    if backend is None:
        # TODO Print rejection reason
        logger.debug("Release was rejected: %s", announcement.title)
    else:
        logger.info("%s approved release: %s", backend.name, announcement.title)
        with db.db_session:
            db_announced = db.get_announcement(db_announced.id)
            db.insert_snatched(db_announced, backend.name)


async def on_message(tracker, source, target, message):
    announcement = _check_and_parse(tracker, source, target, message)
    if announcement is None:
        return

    try:
        await _handle_announcement(tracker, announcement)
    except TransactionError:
        logger.exception("Database transaction failed for %s", announcement.title)



import logging
import re
from asyncio import Lock

from arrnounced.session_provider import SessionProvider


logger = logging.getLogger("BACKEND")


def _extract_approval(json_response, backend_name):
    try:
        return json_response["approved"]
    except TypeError:
        logger.warning("No valid JSON reply from %s", backend_name, exc_info=True)
    except KeyError:
        logger.warning("No approval info in reply from %s", backend_name, exc_info=True)
    return False


class Backend:
    def __init__(self, user_backend):
        self.apikey = user_backend.apikey
        self.url = user_backend.url
        self.name = user_backend.name
        self.lock = Lock()

    def _create_json(self, announcement):
        params = {
            "title": announcement.title,
            "downloadUrl": announcement.torrent_url,
            "protocol": "Torrent",
            "publishDate": announcement.date.isoformat(),
        }

        return params

    async def _send_notification(self, announcement):
        # Mitigate https://github.com/Sonarr/Sonarr/issues/2975
        async with Lock():
            json_response = await SessionProvider.post(
                url=f"{self.url}{self.api_path}",
                headers={"X-Api-Key": self.apikey},
                json=self._create_json(announcement),
            )

        return json_response

    async def notify(self, announcement):
        json_response = await self._send_notification(announcement)
        if not json_response:
            return False
        return _extract_approval(json_response, self.name)


class UseIndexer(Backend):
    def _create_json(self, announcement):
        params = super()._create_json(announcement)
        params["indexer"] = "Irc" + announcement.indexer

        return params


class Sonarr(UseIndexer):
    api_path = "/api/release/push"


class Radarr(UseIndexer):
    api_path = "/api/release/push"


class Lidarr(Backend):
    api_path = "/api/v1/release/push"


backend_mapping = {
    "sonarr": Sonarr,
    "radarr": Radarr,
    "lidarr": Lidarr,
}


_backends = {}


def init(user_backends):
    for user_backend in user_backends:
        _backends[user_backend.name] = backend_mapping[user_backend.type](user_backend)


async def stop():
    await SessionProvider.close_session()


def get_configured_backends():
    return list(_backends.keys())


def notify_which_backends(tracker, announced_category):
    notify_backends = [_backends[b] for b in tracker.config.always_notify_backends]

    if announced_category:
        for (
            backend_name,
            category_regex,
        ) in tracker.config.category_notify_backends.items():
            if re.search(category_regex, announced_category, re.IGNORECASE):
                notify_backends.append(_backends[backend_name])

    # Return all configured backends if none where specified
    if (
        len(tracker.config.always_notify_backends) == 0
        and len(tracker.config.category_notify_backends) == 0
    ):
        notify_backends = list(_backends.values())

    return notify_backends


async def notify(announcement, backends):
    for backend in backends:
        if await backend.notify(announcement):
            return backend

    return None


def get_backend(backend_name):
    return _backends.get(backend_name)


async def renotify(announcement, backend):
    return await backend.notify(announcement)



#!/usr/bin/env python3
import argparse
import logging
import os
import sys
from pathlib import Path

from arrnounced import __version__
from arrnounced import backend
from arrnounced import config
from arrnounced import db
from arrnounced import log
from arrnounced import manager


def _is_file(path):
    if not os.path.isfile(path):
        print("Error: '" + path + "' is not a valid file")
        return False
    return True


def _is_dir(path):
    if not os.path.isdir(path):
        print("Error: '" + path + "' is not a valid directory")
        return False
    return True


def _validate_args(args):
    checks = [_is_dir(args.data), _is_dir(args.trackers), _is_file(args.config)]
    if len(checks) > sum(checks):
        sys.exit(1)


def main():
    parser = argparse.ArgumentParser(
        description="Arrnounced - Listen for IRC announcements"
    )
    parser.add_argument(
        "-d",
        "--data",
        help="Data directory for storing logs and database. Default ~/.arrnounced",
        default=str(Path.home().joinpath(".arrnounced")),
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Configuration file. Default ~/.arrnounced/settings.toml",
        default=str(Path.home().joinpath(".arrnounced", "settings.toml")),
    )
    parser.add_argument(
        "-t",
        "--trackers",
        help="XML tracker config path. Default ~/.arrnounced/autodl-trackers/trackers",
        default=str(Path.home().joinpath(".arrnounced", "autodl-trackers", "trackers")),
    )
    parser.add_argument("-v", "--verbose", help="Verbose logging", action="store_true")
    parser.add_argument("--version", help="Print version", action="store_true")

    try:
        args = parser.parse_args()
    except Exception as e:
        print(e, file=sys.stderr)
        if isinstance(e, FileNotFoundError):
            config.toml_notice()
        sys.exit(1)

    if args.version:
        print("Arrnounced version", __version__)
        sys.exit(0)

    _validate_args(args)

    user_config = config.init(args.config)
    if user_config is None:
        sys.exit(1)

    log_level = logging.INFO
    if args.verbose or bool(os.getenv("VERBOSE")):
        log_level = logging.DEBUG

    log_file = Path(args.data).joinpath("arrnounced.log")
    log.init_logging(user_config, log_level, log_file)

    if not user_config.validate_config():
        print("Error: Configuration not valid", file=sys.stderr)
        sys.exit(1)

    backend.init(user_config.backends)
    if not db.init(args.data):
        sys.exit(1)

    manager.run(user_config, args.trackers)


if __name__ == "__main__":
    main()



class Tracker:
    def __init__(self, tracker_config):
        self.config = tracker_config
        self.status = TrackerStatus()

    @property
    def name(self):
        return self.config.long_name

    @property
    def type(self):
        return self.config.type


class TrackerStatus:
    def __init__(self):
        self.connected = False
        self.channels = []
        self.latest_announcement = None


class TrackerConfig:
    def __init__(self, user_tracker, xml_config):
        self._xml_config = xml_config
        self._user_tracker = user_tracker.tracker

        self._always_backends = (
            [b.strip() for b in self._user_tracker.get("notify_backends").split(",")]
            if self._user_tracker.get("notify_backends")
            else []
        )

    def setting(self, key):
        return self._user_tracker["settings"].get(key)

    @property
    def irc_port(self):
        return int(self._user_tracker["irc_port"])

    @property
    def irc_nickname(self):
        return self._user_tracker["irc_nickname"]

    @property
    def irc_server(self):
        return str(self._user_tracker["irc_server"])

    @property
    def irc_tls(self):
        return self._user_tracker["irc_tls"]

    @property
    def irc_tls_verify(self):
        return self._user_tracker["irc_tls_verify"]

    @property
    def irc_ident_password(self):
        return self._user_tracker.get("irc_ident_password")

    @property
    def irc_inviter(self):
        return self._user_tracker.get("irc_inviter")

    @property
    def irc_invite_cmd(self):
        return self._user_tracker.get("irc_invite_cmd")

    @property
    def torrent_https(self):
        return self._user_tracker["torrent_https"]

    @property
    def announce_delay(self):
        return self._user_tracker["announce_delay"]

    @property
    def always_notify_backends(self):
        return self._always_backends

    @property
    def category_notify_backends(self):
        return self._user_tracker["category"]

    @property
    def short_name(self):
        return self._xml_config.tracker_info["shortName"]

    @property
    def long_name(self):
        return self._xml_config.tracker_info["longName"]

    @property
    def type(self):
        return self._xml_config.tracker_info["type"]

    @property
    def user_channels(self):
        return [
            x.strip().lower() for x in self._user_tracker["irc_channels"].split(",")
        ]

    # Return both channels from XML and user config
    @property
    def irc_channels(self):
        for server in self._xml_config.servers:
            for channel in server.channels:
                yield channel

        for channel in self.user_channels:
            yield channel

    @property
    def announcer_names(self):
        for server in self._xml_config.servers:
            for announcer in server.announcers:
                yield announcer

    @property
    def line_patterns(self):
        return self._xml_config.line_patterns

    @property
    def multiline_patterns(self):
        return self._xml_config.multiline_patterns

    @property
    def ignores(self):
        return self._xml_config.ignores

    @property
    def line_matched(self):
        return self._xml_config.line_matched



import inspect
import logging
import os
from flask import Flask
from flask import jsonify
from flask import redirect
from flask import render_template
from flask import request
from flask import send_from_directory
from flask import url_for
from flask_login import LoginManager
from flask_login import login_user, logout_user
from flask_login import login_required
from flask_login import UserMixin
from pathlib import Path

from arrnounced import log
from arrnounced import web_handler

logger = logging.getLogger("WEB-UI")
table_row_count = 20


class User(UserMixin):
    def get_id(self):
        return 1


# Template directory is the current ([0]) stack object's ([1]) directory + templates
templates_dir = Path(os.path.dirname(os.path.abspath(inspect.stack()[0][1]))).joinpath(
    "templates"
)
app = Flask("Arrnounced", template_folder=templates_dir)
# This will invalidate logins on each restart of Arrnounced
# But I'm too lazy to think of something else at the moment
app.secret_key = os.urandom(16)

login_manager = LoginManager(app=app)
login_manager.login_view = "login"
user = User()
user_config = None


def run(config):
    global user_config
    user_config = config
    try:
        app.run(
            debug=False,
            host=user_config.webui_host,
            port=user_config.webui_port,
            use_reloader=False,
        )
    except OSError as e:
        logger.error("Error starting webserver: %s", e)


@app.route("/shutdown", methods=["GET", "POST"])
def shutdown():
    if not user_config.webui_shutdown:
        return redirect(url_for("index"))

    logger.info("Shutting down Arrnounced")
    logger.info("Disable shutdown by removing webui.shutdown from config")
    web_handler.shutdown()
    func = request.environ.get("werkzeug.server.shutdown")
    if func is None:
        raise RuntimeError("Not running with the Werkzeug Server")
    func()
    return "Shutting down..."


@login_manager.user_loader
def load_user(id):
    return user


@app.route("/login", methods=["GET", "POST"])
def login():
    error = None
    if request.method == "POST" or not user_config.login_required:
        if user_config.login(
            request.form.get("username"), request.form.get("password")
        ):
            login_user(user)
            return redirect(url_for("index"))
        else:
            error = "Invalid credentials"
    return render_template("login.html", error=error)


@app.route("/logout", methods=["GET", "POST"])
def logout():
    logout_user()
    return redirect(url_for("login"))


@app.route("/assets/<path:path>")
def send_asset(path):
    return send_from_directory(
        str(templates_dir) + "/assets/{}".format(os.path.dirname(path)),
        os.path.basename(path),
    )


@app.route("/")
@login_required
def index():
    announce_pages, snatch_pages = web_handler.get_page_counts(table_row_count)
    return render_template(
        "index.html",
        announcement_pages=announce_pages,
        snatch_pages=snatch_pages,
        login_required=user_config.login_required,
    )


@app.route("/logs")
@login_required
def logs():
    logs = []
    for log_line in log.get_logs():
        logs.append({"time": log_line[0], "tag": log_line[1], "msg": log_line[2]})

    return render_template(
        "logs.html", logs=logs, login_required=user_config.login_required
    )


# TODO: Reintroduce check ability
@app.route("/<pvr_name>/check", methods=["POST"])
@login_required
def check(pvr_name):
    try:
        data = request.json
        if "apikey" in data and "url" in data:
            # Check if api key is valid
            logger.debug(
                "Checking whether apikey: %s is valid for: %s",
                data.get("apikey"),
                data.get("url"),
            )

            resp = {}
            # headers = {"X-Api-Key": data.get("apikey")}
            # resp = requests.get(
            #     url="{}/api/diskspace".format(data.get("url")), headers=headers
            # ).json()
            logger.debug("check response: %s", resp)

            if "error" not in resp:
                return "OK"

    # TODO: Catch more specific types
    except Exception:
        logger.exception("Exception while checking " + pvr_name + " apikey:")

    return "ERR"


@app.route("/notify", methods=["POST"])
@login_required
def notify():
    data = request.json
    if "id" in data and "backend_name" in data:
        if web_handler.notify_backend(data["id"], data["backend_name"]):
            return "OK"
    else:
        logger.warning("Missing data in notify request")

    return "ERR"


@app.route("/announced", methods=["POST"])
@login_required
def announced():
    page_nr = 1
    if "page_nr" in request.json:
        page_nr = request.json["page_nr"]

    announced_page, configured_backends = web_handler.get_announced_page(
        page_nr, table_row_count
    )
    announced = jsonify(
        announces=announced_page,
        backends=configured_backends,
    )
    return announced


@app.route("/snatched", methods=["POST"])
@login_required
def snatched():
    page_nr = 1
    if "page_nr" in request.json:
        page_nr = request.json["page_nr"]

    snatched = jsonify(snatches=web_handler.get_snatched_page(page_nr, table_row_count))
    return snatched



import logging
import os
import re
import defusedxml.ElementTree as ET
from collections import namedtuple

from arrnounced.announcement import (
    Var,
    Http,
    Extract,
    ExtractOne,
    ExtractTags,
    VarReplace,
    SetRegex,
    If,
)


# TODO: compile and store regex
logger = logging.getLogger("TRACKER_CONF")
debug = False

Server = namedtuple("Server", "names channels announcers")
Ignore = namedtuple("Ignore", "regex expected")


def get_tracker_xml_configs(tracker_config_path):
    xml_configs = {}
    for tracker_file in sorted(os.listdir(tracker_config_path)):
        tree = ET.parse(tracker_config_path + "/" + tracker_file)
        tracker = TrackerXmlConfig()

        try:
            if tracker.parse_config(tree.getroot()):
                xml_configs[tracker.tracker_info["type"]] = tracker
            else:
                logger.error("Could not parse tracker XML config '%s'", tracker_file)
        except KeyError as e:
            logger.error(
                "Could not parse tracker XML config '%s', missing attribute %s",
                tracker_file,
                e,
            )

    return xml_configs


def _is_optional(element):
    return (
        False
        if "optional" not in element.attrib
        else element.attrib["optional"] == "true"
    )


def var_creator(element):
    return Var(element.attrib["name"], element.findall("./*"))


def http_creator(element):
    return Http()


def extract_creator(element):
    groups = element.findall("./vars/*")
    groupList = []
    for group in groups:
        groupList.append(group.attrib["name"])

    return Extract(
        element.attrib["srcvar"] if "srcvar" in element.attrib else None,
        element.find("./regex").attrib["value"],
        groupList,
        _is_optional(element),
    )


def extract_one_creator(element):
    extracts = []
    xml_extracts = element.findall("./extract")
    for xml_extract in xml_extracts:
        extracts.append(extract_creator(xml_extract))
    return ExtractOne(extracts)


def extract_tags_creator(element):
    srcvar = element.attrib["srcvar"]
    split = element.attrib["split"]

    setvarifs = []
    for setvarif in element.findall("./setvarif"):
        setvarifs.append(
            ExtractTags.SetVarIf(
                setvarif.attrib["varName"],
                setvarif.attrib.get("regex"),
                setvarif.attrib.get("value"),
                setvarif.attrib.get("newValue"),
            )
        )

    return ExtractTags(srcvar, split, setvarifs)


def var_replace_creator(element):
    return VarReplace(
        element.attrib["name"],
        element.attrib["srcvar"],
        element.attrib["regex"],
        element.attrib["replace"],
    )


def set_regex_creator(element):
    return SetRegex(
        element.attrib["srcvar"],
        element.attrib["regex"],
        element.attrib["varName"],
        element.attrib["newValue"],
    )


def if_creator(element):
    line_matches = []
    for matched in element.findall("./*"):
        line_matches.append(_line_match_creators[matched.tag](matched))
    return If(element.attrib["srcvar"], element.attrib["regex"], line_matches)


_line_match_creators = {
    "var": var_creator,
    "http": http_creator,
    "extract": extract_creator,
    "extractone": extract_one_creator,
    "extracttags": extract_tags_creator,
    "varreplace": var_replace_creator,
    "setregex": set_regex_creator,
    "if": if_creator,
}


class TrackerXmlConfig:
    def __init__(self):
        self.settings = []
        self.servers = []
        self.line_matched = []
        self.line_patterns = []
        self.multiline_patterns = []
        self.ignores = []

    def parse_config(self, root):  # noqa: C901
        self.tracker_info = root.attrib

        for setting in root.findall("./settings/*"):
            if "description" not in setting.tag:
                self.settings.append(re.sub("^(gazelle_|cookie_)", "", setting.tag))

        for server in root.findall("./servers/*"):
            self.servers.append(
                Server(
                    server.attrib["serverNames"].lower().split(","),
                    server.attrib["channelNames"].lower().split(","),
                    server.attrib["announcerNames"].split(","),
                )
            )

        for extract in root.findall("./parseinfo/linepatterns/*"):
            self.line_patterns.append(extract_creator(extract))

        for extract in root.findall("./parseinfo/multilinepatterns/*"):
            self.multiline_patterns.append(extract_creator(extract))

        for element in root.findall("./parseinfo/linematched/*"):
            self.line_matched.append(_line_match_creators[element.tag](element))

        for ignore in root.findall("./parseinfo/ignore/*"):
            self.ignores.append(
                Ignore(
                    ignore.attrib["value"],
                    (
                        "expected" not in ignore.attrib
                        or ignore.attrib["expected"] == "true"
                    ),
                )
            )

        if debug:
            for info in self.tracker_info:
                print(info, root.attrib[info])
            print("\tSettings")
            for setting in self.settings:
                print("\t\t", setting)
            print("\tServer")
            for server in self.servers:
                print("\t\tnames:", server.names)
                print("\t\tchannels", server.channels)
                print("\t\tannouncers", server.announcers)
            print("\tTorrentUrl")
            for var in self.line_matched:
                print("\t\t", var.varType, ": ", var.name)
            print("\tLinePatterns")
            for pattern in self.line_patterns:
                print("\t\t", pattern.regex, pattern.optional)
                for group in pattern.groups:
                    print("\t\t\t", group)
            print("\tMultiLinePattern")
            for pattern in self.multiline_patterns:
                print("\t\t", pattern.regex, pattern.optional)
                for group in pattern.groups:
                    print("\t\t\t", group)
            print("\tIgnores")
            for ignore in self.ignores:
                print("\t\t", ignore.regex, "-", ignore.expected)

        if self.tracker_info is None:
            logger.error("No 'tracker_info' found")
            return False
        elif 0 == len(self.line_patterns) and 0 == len(self.multiline_patterns):
            logger.error("No announcement patterns found")
        elif len(self.servers) == 0:
            logger.error("No servers found")
            return False
        elif len(self.line_matched) == 0:
            logger.error("No items to build URL")
            return False

        return True



import time
import logging
import re

from enum import Enum
from itertools import filterfalse
from multiprocessing import Lock

from arrnounced.utils import get_default_variables

logger = logging.getLogger("ANNOUNCE_PARSER")


class ParseStatus(Enum):
    MATCH = 1
    NO_MATCH = 2
    CONTINUE = 3


def parse(tracker, message):
    parse_status = ParseStatus.NO_MATCH
    pattern_groups = {}
    if len(tracker.config.line_patterns) > 0:
        parse_status, pattern_groups = _parse_singleline_patterns(
            tracker.config.line_patterns, message
        )
    elif len(tracker.config.multiline_patterns) > 0:
        parse_status, pattern_groups = _parse_multiline_patterns(tracker, message)

    if not _is_parsing_ok(tracker, parse_status, message):
        return None

    return pattern_groups


def _is_parsing_ok(tracker, parse_status, message):
    if parse_status == ParseStatus.NO_MATCH:
        if _ignore_message(tracker.config.ignores, message):
            logger.debug("%s: Message ignored: %s", tracker.config.short_name, message)
        else:
            logger.warning(
                "%s: No match found for '%s'", tracker.config.short_name, message
            )
    elif parse_status == ParseStatus.CONTINUE:
        logger.debug(
            "%s: Messages in announcement still remaining",
            tracker.config.short_name,
        )

    return parse_status == ParseStatus.MATCH


def _ignore_message(ignores, message):
    for ignore in ignores:
        # If message matches an expected regex it will be ignord
        # If it's not expected it works as an inverse, ignore everything which doesn't match.
        if bool(re.search(ignore.regex, message)) == ignore.expected:
            return True
    return False


def _parse_singleline_patterns(line_patterns, message):
    index, pattern_groups = _parse_message(line_patterns, message)
    if index == -1:
        return ParseStatus.NO_MATCH, {}
    else:
        variables = get_default_variables()
        variables.update(pattern_groups)
        return ParseStatus.MATCH, variables


def _parse_message(pattern_list, message):
    for i, pattern in enumerate(pattern_list, start=0):
        match_groups = pattern.process_string(message)
        if match_groups is not None:
            return i, match_groups
    return -1, {}


######################
# Multi line patterns
######################

multiline_matches = {}
mutex = Lock()


class MultilineMatch:
    def __init__(self):
        self.time = time.time()
        self.pattern_groups = get_default_variables()
        self.matched_index = -1

    # Returns true if more than 15 seconds has passed since instantiated
    def too_old(self):
        return (time.time() - self.time) > 15


# Returning None means the message matched but still waiting for remaning messages.
# Returning an empty dictionary means the message did not match anything.
def _parse_multiline_patterns(tracker, message):
    logger.debug(
        "%s: Parsing multiline annoucement '%s'", tracker.config.short_name, message
    )
    match_index, match_groups = _parse_message(
        tracker.config.multiline_patterns, message
    )

    if match_index == -1:
        return ParseStatus.NO_MATCH, {}

    is_last_pattern = _is_last_multiline_pattern(
        tracker.config.multiline_patterns, match_index
    )

    with mutex:
        multiline_match = _get_multiline_match(
            tracker.config.type,
            tracker.config.multiline_patterns,
            match_index,
            is_last_pattern,
        )

        if multiline_match is None:
            return ParseStatus.NO_MATCH, {}

        multiline_match.matched_index = match_index
        multiline_match.pattern_groups.update(match_groups)

    if is_last_pattern:
        return ParseStatus.MATCH, multiline_match.pattern_groups

    return ParseStatus.CONTINUE, {}


def _is_valid_next_index(matched_index, next_index, patterns):
    return (next_index - matched_index) == 1 or (  # Next match
        next_index - matched_index > 1
        and all(  # Or optionals in between
            pattern.optional for pattern in patterns[matched_index + 1 : next_index]
        )
    )


# Returns True if match_index is the last pattern in the announcement
# OR if the remaining patterns are optional
def _is_last_multiline_pattern(multiline_patterns, match_index):
    return match_index + 1 == len(multiline_patterns) or all(
        pattern.optional for pattern in multiline_patterns[match_index + 1 :]
    )


def _get_multiline_match(tracker_type, patterns, match_index, last_pattern):
    global multiline_matches
    if tracker_type not in multiline_matches:
        multiline_matches[tracker_type] = []

    _clean_old_multi_announcements(multiline_matches[tracker_type])

    if match_index == 0:
        multiline_match = MultilineMatch()
        multiline_matches[tracker_type].append(multiline_match)
        return multiline_match

    for i, multiline_match in enumerate(multiline_matches[tracker_type]):
        if _is_valid_next_index(multiline_match.matched_index, match_index, patterns):
            if last_pattern:
                del multiline_matches[tracker_type][i]
            return multiline_match

    return None


def _clean_old_multi_announcements(matches):
    removes = list(filterfalse(lambda x: not x.too_old(), matches))
    for remove in removes:
        logger.warning(
            "Announcement is too old, discarding: %s",
            list(remove.pattern_groups.values()),
        )
        matches.remove(remove)



import datetime
import logging
import re


logger = logging.getLogger("UTILS")


def strip_irc_color_codes(line):
    line = re.sub(r"\x03\d\d?,\d\d?", "", line)
    line = re.sub(r"\x03\d\d?", "", line)
    line = re.sub(r"[\x01-\x1F]", "", line)
    return line


def replace_spaces(text, new):
    return re.sub("[ ]{1,}", new, text)


# credits: http://code.activestate.com/recipes/576880-convert-datetime-in-python-to-user-friendly-repres/
def human_datetime(date_time):
    """
    converts a python datetime object to the
    format "X days, Y hours ago"

    @param date_time: Python datetime object

    @return:
        fancy datetime:: string
    """
    current_datetime = datetime.datetime.now()
    delta = str(current_datetime - date_time)
    if delta.find(",") > 0:
        days, hours = delta.split(",")
        days = int(days.split()[0].strip())
        hours, minutes = hours.split(":")[0:2]
    else:
        hours, minutes = delta.split(":")[0:2]
        days = 0
    days, hours, minutes = int(days), int(hours), int(minutes)
    datelets = []
    years, months, xdays = None, None, None

    def plural(x):
        return "s" if x != 1 else ""

    if days >= 365:
        years = int(days / 365)
        datelets.append("%d year%s" % (years, plural(years)))
        days = days % 365
    if days >= 30 and days < 365:
        months = int(days / 30)
        datelets.append("%d month%s" % (months, plural(months)))
        days = days % 30
    if not years and days > 0 and days < 30:
        xdays = days
        datelets.append("%d day%s" % (xdays, plural(xdays)))
    if not (months or years) and hours != 0:
        datelets.append("%d hour%s" % (hours, plural(hours)))
    if not (xdays or months or years):
        datelets.append("%d minute%s" % (minutes, plural(minutes)))
    return ", ".join(datelets) + " ago."


def get_default_variables():
    variables = {
        "releaseType": "",
        "freeleech": "",
        "freeleechPercent": "",
        "origin": "",
        "releaseGroup": "",
        "category": "",
        "torrentName": "",
        "uploader": "",
        "torrentSize": "",
        "preTime": "",
        "torrentUrl": "",
        "torrentSslUrl": "",
        "year": "",
        "name1": "",  # artist, show, movie
        "name2": "",  # album
        "season": "",
        "episode": "",
        "resolution": "",
        "source": "",
        "encoder": "",
        "container": "",
        "format": "",
        "bitrate": "",
        "media": "",
        "tags": "",
        "scene": "",
        "log": "",
        "logScore": "",
        "cue": "",
    }
    return variables



import logging
import urllib.parse
from datetime import datetime
from enum import Enum
import re

logger = logging.getLogger("ANNOUNCEMENT")
debug = False


class Announcement:
    def __init__(self, title, url, category=None, date=None, indexer=None):
        self.title = title
        self.torrent_url = url
        self.category = category
        self.date = date
        self.indexer = indexer


def create_announcement(tracker, variables):
    for line_match in tracker.config.line_matched:
        line_match.process(tracker.config, variables)

    _insert_ssl_url(variables)

    torrent_url_var = "torrentSslUrl" if tracker.config.torrent_https else "torrentUrl"
    if not variables.get("torrentName"):
        logger.warning("Missing torrent name")
        return None
    elif not variables.get(torrent_url_var):
        logger.warning("Missing torrent URL")
        return None

    return Announcement(
        variables["torrentName"],
        variables[torrent_url_var],
        variables.get("category"),
        date=datetime.now(),
        indexer=tracker.config.short_name,
    )


def _insert_ssl_url(variables):
    if variables.get("torrentUrl") and not variables.get("torrentSslUrl"):
        variables["torrentSslUrl"] = re.sub("^https?", "https", variables["torrentUrl"])


class Var:
    def __init__(self, var_name, elements):
        self.var_name = var_name
        self.elements = []
        for element in elements:
            self.elements.append(Var.Element(element.tag, element.attrib))

    def process(self, tracker_config, variables):
        var = ""
        for element in self.elements:
            var_shard = None
            if element.var_type is self.Element.Type.STRING:
                var_shard = element.name
            else:
                if element.name in variables:
                    var_shard = variables[element.name]
                else:
                    var_shard = tracker_config.setting(element.name)

                if element.var_type is self.Element.Type.VARENC:
                    var_shard = urllib.parse.quote_plus(var_shard)

            if not var_shard:
                logger.warning(
                    "Could not build var '%s', missing variable '%s'",
                    self.var_name,
                    element.name,
                )
                return

            var = var + var_shard

        if debug:
            logger.debug("Setting variable: %s = %s", self.var_name, var)
        variables[self.var_name] = var

    class Element:
        class Type(Enum):
            STRING = 1
            VAR = 2
            VARENC = 3

        type_to_name = {Type.STRING: "value", Type.VAR: "name", Type.VARENC: "name"}

        def __init__(self, var_type, var):
            self.var_type = self.Type[var_type.upper()]
            self.name = var[self.type_to_name[self.var_type]]


class Http:
    def process(self, tracker_config, variables):
        logger.warning(
            "HTTP header (e.g. cookie) in tracker configuration not supported. "
            + "This might cause problems downloading the torrent file."
        )


class Extract:
    def __init__(self, srcvar, regex, groups, optional):
        self.srcvar = srcvar
        self.regex = regex
        self.groups = groups
        self.optional = optional

    # Returns None when no match was found
    def process_string(self, string):
        matches = re.search(self.regex, string)
        if matches:
            match_groups = {}
            for j, group_name in enumerate(self.groups, start=1):
                # Filter out missing non-capturing groups
                match = matches.group(j)
                if match is not None and not match.isspace():
                    match_groups[group_name] = match.strip()
            return match_groups

        return None

    def get_extract_variables(self, variables):
        match_groups = None
        if self.srcvar in variables:
            match_groups = self.process_string(variables[self.srcvar])
        return match_groups

    def process(self, tracker_config, variables):
        match_groups = self.get_extract_variables(variables)

        if match_groups is not None:
            if debug:
                for k in match_groups:
                    logger.debug("Setting variable: %s = %s", k, match_groups[k])
            variables.update(match_groups)
        elif not self.optional:
            logger.warning(
                "Extract: Variable '%s' did not match regex '%s'",
                self.srcvar,
                self.regex,
            )


class ExtractOne:
    def __init__(self, extracts):
        self.extracts = extracts

    def process(self, tracker_config, variables):
        for extract in self.extracts:
            extract_vars = extract.get_extract_variables(variables)
            if extract_vars is not None:
                if debug:
                    for k in extract_vars:
                        logger.debug("Setting variable: %s = %s", k, extract_vars[k])
                variables.update(extract_vars)
                return

        logger.warning("ExtractOne: No matching regex found")


class ExtractTags:
    def __init__(self, srcvar, split, setvarifs):
        self.srcvar = srcvar
        self.split = split
        self.setvarifs = setvarifs

    def process(self, tracker_config, variables):
        if self.srcvar not in variables:
            logger.warning(
                "ExtractTags: Could not extract tags, variable '%s' not found",
                self.srcvar,
            )
            return

        for tag_name in [
            x.strip() for x in re.split(self.split, variables[self.srcvar])
        ]:
            if not tag_name:
                continue

            for setvarif in self.setvarifs:
                new_value = setvarif.get_value(tag_name)
                if new_value is not None:
                    if debug:
                        logger.debug(
                            "Setting variable: %s = %s", setvarif.var_name, new_value
                        )
                    variables[setvarif.var_name] = new_value
                    break

    class SetVarIf:
        def __init__(self, var_name, regex, value, new_value):
            self.var_name = var_name
            self.regex = regex
            self.value = value
            self.new_value = new_value

        def get_value(self, tag_name):
            if (self.value is not None and self.value.lower() != tag_name.lower()) or (
                self.regex is not None and re.search(self.regex, tag_name) is None
            ):
                return None

            return self.new_value if self.new_value is not None else tag_name


class VarReplace:
    def __init__(self, name, srcvar, regex, replace):
        self.name = name
        self.srcvar = srcvar
        self.regex = regex
        self.replace = replace

    def process(self, tracker_config, variables):
        if self.srcvar not in variables:
            logger.warning(
                "VarReplace: Could not replace, variable '%s' not found", self.srcvar
            )
            return

        if debug:
            logger.debug(
                "Setting variable: %s = %s",
                self.name,
                re.sub(self.regex, self.replace, variables[self.srcvar]),
            )
        variables[self.name] = re.sub(self.regex, self.replace, variables[self.srcvar])


class SetRegex:
    def __init__(self, srcvar, regex, var_name, new_value):
        self.srcvar = srcvar
        self.regex = regex
        self.var_name = var_name
        self.new_value = new_value

    def process(self, tracker_config, variables):
        if self.srcvar not in variables:
            logger.warning(
                "SetRegex: Could not set variable, '%s' not found", self.srcvar
            )
            return

        if re.search(self.regex, variables[self.srcvar]):
            if debug:
                logger.debug("Setting variable: %s = %s", self.var_name, self.new_value)
            variables[self.var_name] = self.new_value


class If:
    def __init__(self, srcvar, regex, line_matches):
        self.srcvar = srcvar
        self.regex = regex
        self.line_matches = line_matches

    def process(self, tracker_config, variables):
        if self.srcvar not in variables:
            logger.warning(
                "If: Could not check condition, variable '%s' not found", self.srcvar
            )
            return

        if re.search(self.regex, variables[self.srcvar]):
            for matched in self.line_matches:
                matched.process(tracker_config, variables)



from aiohttp import ClientSession, ClientError, ContentTypeError
import logging

logger = logging.getLogger("SESSION")


class SessionProvider:
    session = None

    @classmethod
    def get_session(cls):
        if not cls.session:
            cls.session = ClientSession(raise_for_status=True)
        return cls.session

    @classmethod
    async def close_session(cls):
        if cls.session:
            await cls.session.close()

    @staticmethod
    async def post(url, headers, json):
        try:
            async with SessionProvider.get_session().post(
                url,
                headers=headers,
                json=json,
            ) as http_response:
                return await http_response.json()
        except OSError:
            logger.exception("OS error when pushing release to %s", url)
        except ContentTypeError:
            logger.exception("Invalid JSON response from %s", url)
        except ClientError:
            logger.exception("Client error pushing release to %s", url)

        return None



import logging
import io
import sys
from tomlkit import parse


backend_urls = {
    "sonarr": "http://localhost:8989",
    "radarr": "http://localhost:7878",
    "lidarr": "http://localhost:8686",
}
mandatory_tracker_fields = ["irc_nickname", "irc_server", "irc_port", "irc_channels"]
logger = logging.getLogger("CONFIG")


class UserConfig:
    def __init__(self, toml):
        self.toml = toml

    def validate_config(self):  # noqa: C901
        valid = True

        if bool(self.toml["webui"].get("username")) != bool(
            self.toml["webui"].get("password")
        ):
            logger.error("webui: Must set none or both 'username' and 'password'")
            valid = False

        if len(self.toml["backends"]) == 0:
            logger.error("Must specify at least one backend (Sonarr/Radarr/Lidarr)")
            valid = False

        for section_name, section in self.toml["backends"].items():
            if (
                not section.get("type")
                or section.get("type").lower() not in backend_urls
            ):
                logger.error(
                    "backends.%s: Must specify type, one of sonarr, radarr, lidarr",
                    section_name,
                )
                valid = False
            if not section.get("apikey"):
                logger.error("backends.%s: Must specify apikey", section_name)
                valid = False

        for section_name, section in self.toml["trackers"].items():
            for mandatory in mandatory_tracker_fields:
                if not section.get(mandatory):
                    logger.error("trackers.%s: Must set '%s'", section_name, mandatory)
                    valid = False

            if bool(section.get("irc_inviter")) != bool(section.get("irc_invite_cmd")):
                logger.error(
                    "trackers.%s: Must set both 'irc_inviter' and 'irc_invite_cmd'",
                    section_name,
                )
                valid = False

            always_backends = (
                [b.strip() for b in section.get("notify_backends").split(",")]
                if section.get("notify_backends")
                else []
            )
            category_backends = list(section.get("category").keys())

            for b in always_backends + category_backends:
                if b not in self.toml["backends"]:
                    logger.error(
                        "trackers.%s: No backend named '%s' found", section_name, b
                    )
                    valid = False

            backend_duplicates = [b for b in always_backends if b in category_backends]
            if len(backend_duplicates) != 0:
                logger.error(
                    "trackers.%s: Cannot specify the same backend for both 'notify_backends' and 'category'. Found %s",
                    section_name,
                    ",".join(backend_duplicates),
                )
                valid = False

        valid = _check_empty_values(self.toml, []) and valid
        return valid

    class UserBackend:
        def __init__(self, name, user_backend):
            self.name = name
            self.backend = user_backend

        @property
        def type(self):
            return self.backend["type"].lower()

        @property
        def apikey(self):
            return self.backend["apikey"]

        @property
        def url(self):
            return self.backend.get("url")

    @property
    def backends(self):
        return [UserConfig.UserBackend(n, b) for n, b in self.toml["backends"].items()]

    class UserTracker:
        def __init__(self, tracker_type, user_tracker):
            self.type = tracker_type
            self.tracker = user_tracker

        @property
        def settings(self):
            return self.tracker["settings"]

    @property
    def trackers(self):
        return [
            UserConfig.UserTracker(ttype, t)
            for ttype, t in self.toml["trackers"].items()
        ]

    @property
    def log_to_console(self):
        return self.toml["log"]["to_console"]

    @property
    def log_to_file(self):
        return self.toml["log"]["to_file"]

    @property
    def webui_host(self):
        return str(self.toml["webui"]["host"])

    @property
    def webui_port(self):
        return int(self.toml["webui"]["port"])

    @property
    def webui_shutdown(self):
        return self.toml["webui"]["shutdown"]

    @property
    def login_required(self):
        return self.toml["webui"].get("username") is not None

    def login(self, username, password):
        if self.toml["webui"].get("username") is None:
            return True
        elif (
            self.toml["webui"].get("username") == username
            and self.toml["webui"].get("password") == password
        ):
            return True
        return False


def toml_notice():
    print(
        "Please note that the configuration format has changed to TOML", file=sys.stderr
    )
    print("Because of this", file=sys.stderr)
    print("* The default config file path has changed", file=sys.stderr)
    print("* Config file must be updated to conform with TOML", file=sys.stderr)
    print("See the release notes for more info", file=sys.stderr)


def _init_value(table, keys, value, value_type=None):
    if value_type is None:
        value_type = type(value)

    section = table
    for k in keys[:-1]:
        section = section[k]

    last_key = keys[-1]
    if value is not None and section.get(last_key) is None:
        section[last_key] = value
    elif section.get(last_key) is not None and not isinstance(
        section[last_key], value_type
    ):
        print(
            "Error: {} must be of type {}".format(".".join(keys), value_type.__name__)
        )
        return False

    return True


def _init_backends(toml_cfg):
    valid_types = True
    for backend_name in toml_cfg["backends"]:
        backend_type = toml_cfg["backends"][backend_name].get("type")
        if not backend_type:
            continue
        default_url = backend_urls.get(backend_type.lower())
        if not default_url:
            continue

        # Init default values and check types
        default_backend_values = [
            ("url", default_url, None),
            ("type", None, str),
            ("apikey", None, str),
        ]
        for last_key, default_value, the_type in default_backend_values:
            keys = ["backends", backend_name, last_key]
            valid_types = (
                _init_value(toml_cfg, keys, default_value, the_type) and valid_types
            )
    return valid_types


def _init_trackers(toml_cfg):
    valid_types = True
    for tracker_type in toml_cfg["trackers"]:
        # Init default values
        default_tracker_values = [
            ("irc_tls", False),
            ("irc_tls_verify", False),
            ("torrent_https", False),
            ("announce_delay", 0),
            ("category", {}),
            ("settings", {}),
        ]
        for last_key, default_value in default_tracker_values:
            keys = ["trackers", tracker_type, last_key]
            valid_types = _init_value(toml_cfg, keys, default_value) and valid_types

        # Check mandatory and empty value types
        tracker_types = [
            ("irc_nickname", str),
            ("irc_server", str),
            ("irc_port", int),
            ("irc_channels", str),
            ("irc_ident_password", str),
            ("irc_inviter", str),
            ("irc_invite_cmd", str),
            ("notify_backends", str),
        ]
        for last_key, the_type in tracker_types:
            keys = ["trackers", tracker_type, last_key]
            valid_types = _init_value(toml_cfg, keys, None, the_type) and valid_types

        # All settings must be string
        for setting in toml_cfg["trackers"][tracker_type]["settings"].keys():
            keys = ["trackers", tracker_type, "settings", setting]
            valid_types = _init_value(toml_cfg, keys, None, str) and valid_types

        # All categories must be string
        for category in toml_cfg["trackers"][tracker_type]["category"].keys():
            keys = ["trackers", tracker_type, "category", category]
            valid_types = _init_value(toml_cfg, keys, None, str) and valid_types

    return valid_types


def init(config_path):
    toml_cfg = None
    with io.open(config_path) as f:
        try:
            toml_cfg = parse(f.read())
        except Exception as e:
            print("Error {}: {}".format(config_path, e), file=sys.stderr)
            toml_notice()
            return None

    valid_types = True

    # Init default values
    default_values = [
        (["webui"], {}),
        (["webui", "host"], "0.0.0.0"),  # nosec B104: bind all interfaces
        (["webui", "port"], 3467),
        (["webui", "shutdown"], False),
        (["log"], {}),
        (["log", "to_file"], True),
        (["log", "to_console"], True),
        (["backends"], {}),
        (["trackers"], {}),
    ]
    for keys, default_value in default_values:
        valid_types = _init_value(toml_cfg, keys, default_value) and valid_types

    # Check type of default empty fields
    types = [
        (["webui", "username"], str),
        (["webui", "password"], str),
    ]
    for keys, the_type in types:
        valid_types = _init_value(toml_cfg, keys, None, the_type) and valid_types

    valid_types = _init_backends(toml_cfg) and valid_types
    valid_types = _init_trackers(toml_cfg) and valid_types

    if not valid_types:
        print(
            "Found configuration type error(s), double check your config quote marks, spelling and subsections"
        )
        return None

    # for k, v in toml_cfg.items():
    #    print(k + ": " + str(v))
    return UserConfig(toml_cfg)


def _check_empty_values(section, prior_sections):
    valid = True
    for key, value in section.items():
        if type(value) is dict:
            valid = _check_empty_values(value, prior_sections + [key]) and valid
        elif len(str(value)) == 0:
            logger.error(
                "%s.%s: Empty value in configuration not allowed. Remove instead.",
                ".".join(prior_sections),
                key,
            )
            valid = False
    return valid



import logging
import time

from asyncio import all_tasks, run_coroutine_threadsafe

logger = logging.getLogger("LOOP_UTILS")


class EventLoopUtil:
    def __init__(self):
        self._the_eventloop = None

    def set_eventloop(self, the_eventloop):
        self._the_eventloop = the_eventloop

    def run(self, the_task):
        return run_coroutine_threadsafe(the_task, self._the_eventloop)

    def wait_till_complete(self):
        logger.info("Waiting for eventloop tasks...")
        # The backend stop task has likely not had time to start yet
        # Therefore sleep a while until it does
        time.sleep(1)
        while len(all_tasks(self._the_eventloop)) != 0:
            time.sleep(1)
        logger.info("Tasks done")

    def stop_eventloop(self):
        self._the_eventloop.call_soon_threadsafe(self._the_eventloop.stop)

        logger.info("Stopping eventloop...")
        while self._the_eventloop.is_running():
            time.sleep(1)
        self._the_eventloop.close()
        logger.info("Eventloop stopped")


eventloop_util = EventLoopUtil()



import logging
from math import ceil
from pony.orm.core import TransactionError

from arrnounced import db
from arrnounced import irc
from arrnounced import utils

from arrnounced.backend import renotify, get_configured_backends, get_backend, stop
from arrnounced.eventloop_utils import eventloop_util
from arrnounced.announcement import Announcement

logger = logging.getLogger("WEB-HANDLER")


def shutdown():
    for task in irc.get_stop_tasks():
        eventloop_util.run(task)
    eventloop_util.wait_till_complete()

    eventloop_util.run(stop())
    eventloop_util.wait_till_complete()

    eventloop_util.stop_eventloop()


# Request to check this torrent again
def _locked_notify(announcement_id, backend):
    with db.db_session:
        db_announcement = db.get_announcement(announcement_id)

    if db_announcement is None or len(db_announcement.title) == 0:
        logger.warning("Announcement to notify not found in database")
        return False

    logger.debug("Checking announcement again: %s", db_announcement.title)
    announcement = Announcement(
        db_announcement.title,
        db_announcement.torrent,
        indexer=db_announcement.indexer,
        date=db_announcement.date,
    )

    future = eventloop_util.run(renotify(announcement, backend))
    if future.result():
        logger.debug("%s accepted the torrent this time!", backend.name)
        with db.db_session:
            db_announcement = db.get_announcement(db_announcement.id)
            db.insert_snatched(db_announcement, backend.name)
        return True

    logger.debug("%s still refused this torrent...", backend.name)
    return False


def notify_backend(announcement_id, backend_name):
    backend = get_backend(backend_name)
    if backend:
        try:
            return _locked_notify(announcement_id, backend)
        except TransactionError:
            logger.exception("Database transaction failed")
    else:
        logger.warning(
            "Could not find the requested backend '%s'",
            backend_name,
        )

    return False


def get_page_counts(page_size):
    try:
        with db.db_session:
            return (
                ceil(db.get_announced_count() / page_size),
                ceil(db.get_snatched_count() / page_size),
            )
    except TransactionError as e:
        logger.error("%s: %s", type(e).__name__, e)

    return 0, 0


def get_announced_page(page_nr, page_size):
    try:
        with db.db_session:
            return (
                [
                    e.serialize(utils.human_datetime)
                    for e in db.get_announced(limit=page_size, page=page_nr)
                ],
                get_configured_backends(),
            )
    except TransactionError as e:
        logger.error("%s: %s", type(e).__name__, e)

    return [], []


def get_snatched_page(page_nr, page_size):
    try:
        with db.db_session:
            return [
                db.snatched_to_dict(e, utils.human_datetime)
                for e in db.get_snatched(limit=page_size, page=page_nr)
            ]
    except TransactionError as e:
        logger.error("%s: %s", type(e).__name__, e)

    return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arrnounced']

package_data = \
{'': ['*'],
 'arrnounced': ['templates/*',
                'templates/assets/bootstrap/css/*',
                'templates/assets/bootstrap/fonts/*',
                'templates/assets/bootstrap/js/*',
                'templates/assets/css/*',
                'templates/assets/js/*']}

install_requires = \
['Flask-Login==0.4.1',
 'Flask==1.1.1',
 'aiohttp==3.8.1',
 'defusedxml==0.6.0',
 'pony==0.7.14',
 'pydle==0.9.4',
 'tomlkit==0.7.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['arrnounced = arrnounced.cli:main']}

setup_kwargs = {
    'name': 'arrnounced',
    'version': '0.7.2',
    'description': 'Notify Sonarr/Radarr/Lidarr of tracker IRC announcements',
    'long_description': '# Arrnounced\nNotify Sonarr/Radarr/Lidarr of tracker IRC announcements.\n\nBuilt on the work of\n[sonarrAnnounced](https://github.com/l3uddz/sonarrAnnounced) with tracker\nconfiguration from\n[autodl-trackers](https://github.com/autodl-community/autodl-trackers) (used by\n[autodl-irssi](https://github.com/autodl-community/autodl-irssi))\n\n## Features\n* All trackers from\n[autodl-trackers](https://github.com/autodl-community/autodl-trackers/tree/master/trackers)\nare supported.\n* Web UI to list announcements and accepted notifications\n    * Ability to search among the announcements remains to be implemented though\n* Notify based on announcement category\n* Configurable delay between IRC announcement and notification\n\nOnly a few of the supported trackers are tested at the moment. Please report any issues you find.\n\n# Setup\n\n_Release v0.7 updated the configuration format. See the [release\nnotes](https://github.com/weannounce/arrnounced/releases/tag/v0.7) for more\ninformation._\n\n## Configuration\nThe default configuration path is `~/.arrnounced/settings.toml`.\n[example.cfg](https://github.com/weannounce/arrnounced/blob/master/example.cfg)\nis the acting configuration documentation.\n\nThe default XML tracker configuration path is `~/.arrnounced/autodl-trackers/trackers`\n\n## Installation\n\n```bash\n# Optional virtual environment\n$ python -m venv path/to/venv\n$ source path/to/venv/bin/activate\n\n# Install\n$ pip install arrnounced\n\n# Run\n$ arrnounced\n```\n\nConfiguration files path as well as log and database location may be changed with command line arguments.\n\n\n### Docker\n[Arrnounced on dockerhub](https://hub.docker.com/r/weannounce/arrnounced)\n\n* You must provide `settings.toml` in `/config`. This is also where logs and the database will be stored.\n* To access the web UI using bridged network the webui host in settings.toml must be `0.0.0.0`.\n* As Arrnounced runs as a non-root user by default it is recommended to specify your own user to handle write access to `/config`.\n\n```bash\n# Default example\ndocker run -v /path/to/settings:/config \\\n           --user 1000 \\\n           -p 3467:3467 weannounce/arrnounced:latest\n```\n\nThe docker image comes with a snapshot of XML tracker configurations located under `/trackers`. If you prefer your own version you can mount over it.\n\n```bash\n# Example with custom XML tracker configs and verbose logging\ndocker run -v /path/to/settings:/config \\\n           -v /path/to/autodl-trackers/trackers:/trackers \\\n           -e VERBOSE=Y \\\n           --user 1000 \\\n           -p 3467:3467 weannounce/arrnounced:latest\n```\n\n## Database design update\nThe database design was updated in [v0.3](https://github.com/weannounce/arrnounced/releases/tag/v0.3)\n([ef931ee](https://github.com/weannounce/arrnounced/commit/ef931eef27348f82254d601f96d094a7b9f147bb)).\nIf you used Arrnounced prior to this or used its predecessor you have two options.\n* Convert your old database using [convert_db.py](https://github.com/weannounce/arrnounced/blob/master/convert_db.py)\n* Move the old database file for safe keeping and let Arrnounced create a new file.\n\nThe default path to the database is `~/.arrnounced/brain.db`\n',
    'author': 'WeAnnounce',
    'author_email': 'weannounce@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import datetime
import logging
import os

from pony.orm import Database, desc, pony, Required, Set
from pony.orm import db_session  # noqa: F401

logger = logging.getLogger("DB")
db = Database()


class Announced(db.Entity):
    date = Required(datetime.datetime)
    title = Required(str)
    indexer = Required(str)
    torrent = Required(str)
    backend = Required(str)
    snatched = Set("Snatched")

    def serialize(self, transform_date):
        return {
            "id": self.id,
            "date": transform_date(self.date),
            "title": self.title,
            "indexer": self.indexer,
            "torrent": self.torrent,
            "backend": self.backend,
        }


class Snatched(db.Entity):
    date = Required(datetime.datetime)
    announced = Required(Announced)
    backend = Required(str)


def init(destination_dir):
    try:
        db.bind(
            "sqlite",
            os.path.join(os.path.realpath(destination_dir), "brain.db"),
            create_db=True,
        )
        db.generate_mapping(create_tables=True)
    except pony.orm.dbapiprovider.OperationalError as e:
        logger.error(
            "Could not initiate database: %s",
            e,
        )
        return False

    return True


def snatched_to_dict(snatched, transform_date):
    return {
        "date": transform_date(snatched[1]),
        "indexer": snatched[2],
        "title": snatched[3],
        "backend": snatched[4],
    }


def get_snatched(limit, page):
    # Order by first attribute in tuple i.e. s.id
    ss = (
        pony.orm.select(
            (s.id, s.date, a.indexer, a.title, s.backend)
            for s in Snatched
            for a in s.announced
        )
        .order_by(desc(1))
        .limit(limit, offset=(page - 1) * limit)
    )
    return ss


def get_announced(limit, page):
    return (
        Announced.select()
        .order_by(desc(Announced.id))
        .limit(limit, offset=(page - 1) * limit)
    )


def get_announcement(id):
    return Announced.get(id=id)


def insert_announcement(announcement, backends):
    return Announced(
        date=announcement.date,
        title=announcement.title,
        torrent=announcement.torrent_url,
        indexer=announcement.indexer,
        backend=backends,
    )


def insert_snatched(announcement, backend):
    Snatched(date=datetime.datetime.now(), announced=announcement, backend=backend)


def get_announced_count():
    return pony.orm.count(a for a in Announced)


def get_snatched_count():
    return pony.orm.count(s for s in Snatched)



import logging
import re

from logging.handlers import RotatingFileHandler

log_file = None


def init_logging(user_config, log_level, the_log_file):
    global log_file
    log_file = the_log_file
    logFormatter = logging.Formatter(
        "%(asctime)s - %(levelname)s:%(name)-28s - %(message)s"
    )
    rootLogger = logging.getLogger()
    rootLogger.setLevel(log_level)

    if user_config.log_to_console:
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logFormatter)
        rootLogger.addHandler(consoleHandler)

    if user_config.log_to_file:
        fileHandler = RotatingFileHandler(
            log_file, maxBytes=1024 * 1024 * 5, backupCount=5
        )
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)


# Returns tuple of (time, tag, message).
def get_logs():
    with open(log_file) as f:
        for line in f:
            log_parts = re.search(
                r"(^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3})\s-\s(\S+)\s+-\s(.+)",
                line,
            )
            if log_parts:
                yield log_parts.group(1), log_parts.group(2), log_parts.group(3)



import logging
import pydle

from arrnounced import irc_modes
from arrnounced import message_handler

from arrnounced.eventloop_utils import eventloop_util

logger = logging.getLogger("IRC")


class IRC(irc_modes.ModesFixer):
    RECONNECT_MAX_ATTEMPTS = None

    def __init__(self, tracker, event_loop):
        super().__init__(tracker.config.irc_nickname, eventloop=event_loop)
        self.tracker = tracker

    async def connect(self, *args, **kwargs):
        try:
            await super().connect(*args, **kwargs)
        except OSError as e:
            logger.error("%s: %s", type(e).__name__, e)
            await self.on_disconnect(expected=False)

    # Request channel invite or join channel
    async def attempt_join_channel(self):
        if self.tracker.config.irc_invite_cmd is None:
            for channel in self.tracker.config.user_channels:
                logger.info("Joining %s", channel)
                await self.join(channel)
        else:
            logger.info("%s: Requesting invite", self.tracker.config.short_name)
            await self.message(
                self.tracker.config.irc_inviter, self.tracker.config.irc_invite_cmd
            )

    async def on_connect(self):
        logger.info("Connected to: %s", self.tracker.config.irc_server)
        await super().on_connect()

        if self.tracker.config.irc_ident_password is None:
            await self.attempt_join_channel()
        else:
            logger.info("Identifying with NICKSERV")
            await self.rawmsg(
                "PRIVMSG",
                "NICKSERV",
                "IDENTIFY",
                self.tracker.config.irc_ident_password,
            )

    async def on_raw(self, message):
        await super().on_raw(message)

        if message.command == 221 and "+r" in message._raw:
            logger.info("Identified with NICKSERV (221)")
            await self.attempt_join_channel()

    async def on_raw_900(self, message):
        logger.info("Identified with NICKSERV (900)")
        await self.attempt_join_channel()

    async def on_message(self, target, source, message):
        await message_handler.on_message(self.tracker, source, target.lower(), message)

    async def on_invite(self, channel, by):
        logger.info("%s invited us to join %s", by, channel)
        if channel in self.tracker.config.irc_channels:
            await self.join(channel)
        else:
            logger.warning(
                "Skipping join. %s is not in irc_channels list or specified in XML tracker configuration.",
                channel,
            )


pool = pydle.ClientPool()
clients = []


def get_stop_tasks():
    logger.info("Stopping IRC client(s)")
    global clients
    for client in clients:
        yield client.disconnect(expected=True)


def run(trackers):
    global pool, clients
    eventloop_util.set_eventloop(pool.eventloop)

    for tracker in trackers.values():
        logger.info(
            "Connecting to server: %s:%d %s",
            tracker.config.irc_server,
            tracker.config.irc_port,
            ", ".join(tracker.config.user_channels),
        )

        client = IRC(tracker, pool.eventloop)

        clients.append(client)
        try:
            pool.connect(
                client,
                hostname=tracker.config.irc_server,
                port=tracker.config.irc_port,
                tls=tracker.config.irc_tls,
                tls_verify=tracker.config.irc_tls_verify,
            )
        except Exception:
            logger.exception("Error while connecting to: %s", tracker.config.irc_server)

    try:
        pool.handle_forever()
    except Exception:
        logger.exception("Exception pool.handle_forever:")
